from typing import Any, List, Dict
import tempfile
import tarfile
import pathlib
import urllib.request
import shutil

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


CONDA_STORE_UI_VERSION = "0.1.4"
CONDA_STORE_UI_URL = f"https://registry.npmjs.org/@conda-store/conda-store-ui/-/conda-store-ui-{CONDA_STORE_UI_VERSION}.tgz"
CONDA_STORE_UI_FILES = [
    "main.js",
    "main.css",
    "main.css.map",
    "main.js.map",
    "index.html",
]


class DownloadCondaStoreUIHook(BuildHookInterface):
    def clean(self, versions: List[str]) -> None:
        super().clean(versions)
        destination_directory = (
            pathlib.Path(self.root) / "conda_store_server/server/static/conda-store-ui"
        )
        shutil.rmtree(destination_directory, ignore_errors=True)

    def initialize(self, version: str, build_data: Dict[str, Any]) -> None:
        super().initialize(version, build_data)

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = pathlib.Path(tmp_dir)
            tmp_filename = tmp_dir / "conda-store-ui.tgz"

            print(f"Downloading @conda-store/conda-store-ui={CONDA_STORE_UI_VERSION}")
            with urllib.request.urlopen(CONDA_STORE_UI_URL) as response:
                with tmp_filename.open("wb") as f:
                    f.write(response.read())

            print(f"Extracting @conda-store/conda-store-ui={CONDA_STORE_UI_VERSION}")
            with tarfile.open(tmp_filename, "r:gz") as tar:
                tar.extractall(path=tmp_dir)

            source_directory = tmp_dir / "package/dist"
            destination_directory = (
                pathlib.Path(self.root)
                / "conda_store_server/server/static/conda-store-ui"
            )
            destination_directory.mkdir(parents=True, exist_ok=True)

            print(f"Copying files {CONDA_STORE_UI_FILES}")
            for filename in CONDA_STORE_UI_FILES:
                shutil.copy(
                    source_directory / filename,
                    destination_directory / filename,
                )



"""split conda_package into conda_package and conda_package_build

Revision ID: 16f65805dc8f
Revises: 5ad723de2abd
Create Date: 2022-08-24 12:01:48.461989

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "16f65805dc8f"
down_revision = "5ad723de2abd"
branch_labels = None
depends_on = None


def upgrade():

    # To migrate the data from conda_package, we'll use a temporary table.
    # We'll populate conda_package_tmp with the right data, then delete
    # conda_package and rename conda_package_tmp into conda_package
    op.create_table(
        "conda_package_tmp",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column("license", sa.Text(), nullable=True),
        sa.Column("license_family", sa.Unicode(length=64), nullable=True),
        sa.Column("name", sa.Unicode(length=255), nullable=False),
        sa.Column("version", sa.Unicode(length=64), nullable=False),
        sa.Column("summary", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
    )

    # Creates table conda_package_build
    op.create_table(
        "conda_package_build",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("package_id", sa.Integer(), nullable=True),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column("build", sa.Unicode(length=64), nullable=False),
        sa.Column("build_number", sa.Integer(), nullable=False),
        sa.Column("constrains", sa.JSON(), nullable=True),
        sa.Column("depends", sa.JSON(), nullable=False),
        sa.Column("md5", sa.Unicode(length=255), nullable=False),
        sa.Column("sha256", sa.Unicode(length=64), nullable=False),
        sa.Column("size", sa.BigInteger(), nullable=False),
        sa.Column("subdir", sa.Unicode(length=64), nullable=True),
        sa.Column("timestamp", sa.BigInteger(), nullable=True),
        sa.ForeignKeyConstraint(
            ["package_id"],
            ["conda_package_tmp.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "package_id",
            "subdir",
            "build",
            "build_number",
            "sha256",
            name="_conda_package_build_uc",
        ),
    )

    # Creates index on build field
    op.create_index(
        op.f("ix_conda_package_build_build"),
        "conda_package_build",
        ["build"],
        unique=False,
    )

    # Creates table build_conda_package_build, which makes the link
    # between a conda-store build and its package builds
    op.create_table(
        "build_conda_package_build",
        sa.Column("build_id", sa.Integer(), nullable=False),
        sa.Column("conda_package_build_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["build_id"], ["build.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(
            ["conda_package_build_id"], ["conda_package_build.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("build_id", "conda_package_build_id"),
    )

    op.create_table(
        "solve_conda_package_build",
        sa.Column("solve_id", sa.Integer(), nullable=False),
        sa.Column("conda_package_build_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["conda_package_build_id"], ["conda_package_build.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["solve_id"], ["solve.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("solve_id", "conda_package_build_id"),
    )

    with op.batch_alter_table("conda_package_build") as batch_op:
        batch_op.drop_constraint("_conda_package_build_uc", type_="unique")

        batch_op.create_unique_constraint(
            "_conda_package_build_uc",
            ["channel_id", "package_id", "subdir", "build", "build_number", "sha256"],
        )
        batch_op.create_foreign_key(
            "conda_package_build_channel_id_fkey",
            "conda_channel",
            ["channel_id"],
            ["id"],
        )

    # Migration starts here

    # Step 1 :
    # We insert in conda_package_tmp its right data.
    # We need to make a group by and use MAX() on the non-grouped fields, because
    # For a given (channel_id, name, version), the description or the license might be different.
    # Sometimes, the difference is just a dash in the license name : "Apache 2" -> "Apache-2"
    op.execute(
        """
    INSERT INTO conda_package_tmp (channel_id, license, license_family, name, version, summary, description)
            SELECT channel_id, MAX(license), MAX(license_family), name, version, MAX(summary), MAX(description)
            FROM conda_package
            GROUP BY channel_id, name, version
    """
    )

    # Step 2 : populate conda_package_build with the data from conda_package
    # We make a join on conda_package_tmp because we want the new package id (conda_package_tmp.id)
    # and not the former one (conda_package.id)

    # Due to the specific nature of Postgres' group by, we need to run a different query for it.
    # The problem is that doing a `GROUP BY cp.channel_id, subdir, build, build_number, sha256`,
    # implies aggregating on the other fields, and there's no (simple) way
    # to aggregate properly on the JSON fields like `constraints`.
    if op.get_bind().engine.name == "postgresql":

        op.execute(
            """
                INSERT INTO conda_package_build (md5, constrains, sha256, build_number, timestamp, size, build, subdir, depends, package_id, channel_id)
                    SELECT DISTINCT ON(channel_id, subdir, build, build_number, sha256)
                        cp.md5,
                        cp.constrains,
                        cp.sha256,
                        cp.build_number,
                        cp.timestamp,
                        cp.size,
                        cp.build,
                        cp.subdir,
                        cp.depends,
                        tmp.id,
                        cp.channel_id
                    FROM conda_package cp
                    LEFT JOIN conda_package_tmp tmp
                        ON cp.channel_id = tmp.channel_id
                        AND cp.name = tmp.name
                        AND cp.version = tmp.version
                    ORDER BY channel_id, subdir, build, build_number, sha256;
                """
        )

    else:

        op.execute(
            """
                INSERT INTO conda_package_build (md5, constrains, sha256, build_number, timestamp, size, build, subdir, depends, package_id, channel_id)
                    SELECT  cp.md5,
                            cp.constrains,
                            cp.sha256,
                            cp.build_number,
                            cp.timestamp,
                            cp.size,
                            cp.build,
                            cp.subdir,
                            cp.depends,
                            tmp.id,
                            cp.channel_id
                    FROM conda_package cp
                    LEFT JOIN conda_package_tmp tmp
                        ON cp.channel_id = tmp.channel_id
                        AND cp.name = tmp.name
                        AND cp.version = tmp.version
                    GROUP BY cp.channel_id, subdir, build, build_number, sha256
                    ORDER BY cp.channel_id, subdir, build, build_number, sha256;
                """
        )

    # Step 3 : migrate the packages of builds, to point to conda_package_build data
    # instead of conda_package.
    op.execute(
        """
                INSERT INTO build_conda_package_build (build_id, conda_package_build_id)
                SELECT bcp.build_id, cpb.id
                FROM build_conda_package bcp
                LEFT JOIN conda_package cp ON bcp.conda_package_id = cp.id
                LEFT JOIN conda_package_build cpb ON cp.sha256 = cpb.sha256 AND cp.channel_id = cpb.channel_id;
            """
    )

    # Step 4 : same logic with the solves
    op.execute(
        """
                INSERT INTO solve_conda_package_build (solve_id, conda_package_build_id)
                SELECT scp.solve_id, cpb.id
                FROM solve_conda_package scp
                LEFT JOIN conda_package cp ON scp.conda_package_id = cp.id
                LEFT JOIN conda_package_build cpb ON cp.sha256 = cpb.sha256 AND cp.channel_id = cpb.channel_id;
            """
    )

    # Data are now migrated, let's clean up

    op.drop_table("solve_conda_package")
    op.drop_table("build_conda_package")
    op.drop_table("conda_package")
    op.rename_table("conda_package_tmp", "conda_package")

    with op.batch_alter_table("build") as batch_op:
        batch_op.create_foreign_key(
            "build_specification_id_fkey",
            "specification",
            ["specification_id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            "build_environment_id_fkey", "environment", ["environment_id"], ["id"]
        )

    with op.batch_alter_table("build_artifact") as batch_op:
        batch_op.create_foreign_key(
            "build_artifact_build_id_fkey", "build", ["build_id"], ["id"]
        )

    with op.batch_alter_table("conda_package") as batch_op:
        batch_op.create_unique_constraint(
            "_conda_package_uc", ["channel_id", "name", "version"]
        )

        batch_op.create_index(
            op.f("ix_conda_package_channel_id"),
            ["channel_id"],
            unique=False,
        )
        batch_op.create_index(op.f("ix_conda_package_name"), ["name"], unique=False)
        batch_op.create_index(
            op.f("ix_conda_package_version"), ["version"], unique=False
        )
        batch_op.create_foreign_key(
            "conda_package_channel_id_fkey",
            "conda_channel",
            ["channel_id"],
            ["id"],
        )

    with op.batch_alter_table("environment") as batch_op:
        batch_op.create_unique_constraint(
            "_namespace_name_uc", ["namespace_id", "name"]
        )
        batch_op.create_foreign_key(
            "environment_namespace_id_fkey",
            "namespace",
            ["namespace_id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            "environment_current_build_id_fkey",
            "build",
            ["current_build_id"],
            ["id"],
        )

    with op.batch_alter_table("solve") as batch_op:
        batch_op.create_foreign_key(
            "solve_specification_id_fkey",
            "specification",
            ["specification_id"],
            ["id"],
        )
    # ### end Alembic commands ###


def downgrade():

    with op.batch_alter_table("solve") as batch_op:
        batch_op.drop_constraint("solve_specification_id_fkey", type_="foreignkey")

    with op.batch_alter_table("environment") as batch_op:
        batch_op.drop_constraint(
            "environment_current_build_id_fkey", type_="foreignkey"
        )
        batch_op.drop_constraint("environment_namespace_id_fkey", type_="foreignkey")
        batch_op.drop_constraint("_namespace_name_uc", type_="unique")

    # To migrate the data from conda_package and conda_package_build back into
    # one table for both, we'll use a temporary table.
    # We'll populate conda_package_tmp with the right data, then delete
    # conda_package and conda_package_build, and then rename conda_package_tmp into conda_package

    op.create_table(
        "conda_package_tmp",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column("build", sa.Unicode(length=64), nullable=False),
        sa.Column("build_number", sa.Integer(), nullable=False),
        sa.Column("constrains", sa.JSON(), nullable=True),
        sa.Column("depends", sa.JSON(), nullable=False),
        sa.Column("license", sa.Text(), nullable=True),
        sa.Column("license_family", sa.Unicode(length=64), nullable=True),
        sa.Column("md5", sa.Unicode(length=255), nullable=False),
        sa.Column("name", sa.Unicode(length=255), nullable=False),
        sa.Column("sha256", sa.Unicode(length=64), nullable=False),
        sa.Column("size", sa.BigInteger(), nullable=False),
        sa.Column("subdir", sa.Unicode(length=64), nullable=True),
        sa.Column("timestamp", sa.BigInteger(), nullable=True),
        sa.Column("version", sa.Unicode(length=64), nullable=False),
        sa.Column("summary", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
    )

    with op.batch_alter_table(
        "conda_package",
        table_args=[
            sa.ForeignKeyConstraint(
                ["channel_id"],
                ["conda_channel.id"],
            ),
            sa.UniqueConstraint(
                "channel_id",
                "subdir",
                "name",
                "version",
                "build",
                "build_number",
                "sha256",
                name="_conda_package_uc",
            ),
        ],
    ):
        pass

    op.create_table(
        "build_conda_package",
        sa.Column("build_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column(
            "conda_package_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.PrimaryKeyConstraint(
            "build_id", "conda_package_id", name="build_conda_package_pkey"
        ),
    )
    op.create_table(
        "solve_conda_package",
        sa.Column("solve_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column(
            "conda_package_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.PrimaryKeyConstraint(
            "solve_id", "conda_package_id", name="solve_conda_package_pkey"
        ),
    )

    # Downgrade migration starts here

    # Step 1 :
    # We insert in conda_package_tmp its right data.

    op.execute(
        """INSERT INTO conda_package_tmp (channel_id, build,build_number,constrains,depends, license, license_family,
                                            md5, name, sha256, size,subdir, timestamp, version, summary, description)
            SELECT  cp.channel_id,
                    cpb.build,
                    cpb.build_number,
                    cpb.constrains,
                    cpb.depends,
                    cp.license,
                    cp.license_family,
                    cpb.md5,
                    cp.name,
                    cpb.sha256,
                    cpb.size,
                    cpb.subdir,
                    cpb.timestamp,
                    cp.version,
                    cp.summary,
                    cp.description
            FROM conda_package cp
            LEFT JOIN conda_package_build cpb ON cpb.package_id = cp.id;
        """
    )

    # Step 2 : migrate the packages of builds, to point to conda_package_build data
    # instead of conda_package.
    op.execute(
        """
                INSERT INTO build_conda_package (build_id, conda_package_id)
                SELECT bcpb.build_id,
                       cpt.id
                FROM build_conda_package_build bcpb
                LEFT JOIN conda_package_build cpb ON bcpb.conda_package_build_id = cpb.id
                LEFT JOIN conda_package_tmp cpt ON cpb.sha256 = cpt.sha256
                AND cpt.channel_id = cpb.channel_id;
            """
    )

    # Step 3 : same logic with the solves
    op.execute(
        """
                INSERT INTO solve_conda_package (solve_id, conda_package_id)
                SELECT scpb.solve_id, cpt.id
                FROM solve_conda_package_build scpb
                LEFT JOIN conda_package_build cpb ON scpb.conda_package_build_id = cpb.id
                LEFT JOIN conda_package_tmp cpt ON cpb.sha256 = cpt.sha256
                AND cpt.channel_id = cpb.channel_id;

            """
    )

    with op.batch_alter_table("build_artifact") as batch_op:
        batch_op.drop_constraint("build_artifact_build_id_fkey", type_="foreignkey")

    with op.batch_alter_table("build") as batch_op:
        batch_op.drop_constraint("build_environment_id_fkey", type_="foreignkey")
        batch_op.drop_constraint("build_specification_id_fkey", type_="foreignkey")

    op.drop_table("solve_conda_package_build")
    op.drop_table("build_conda_package_build")
    op.drop_index(
        op.f("ix_conda_package_build_build"), table_name="conda_package_build"
    )
    op.drop_table("conda_package_build")
    op.drop_table("conda_package")
    op.rename_table("conda_package_tmp", "conda_package")

    # ### end Alembic commands ###



import pytest
import datetime

from conda_store_server.server.auth import (
    AuthenticationBackend,
    RBACAuthorizationBackend,
)
from conda_store_server.schema import AuthenticationToken, Permissions


@pytest.mark.parametrize(
    "expression, resource, match",
    [
        ("e*/d*", "example/data", True),
        ("e*/d*", "example/eta", False),
        ("*e*d*/*e", "deed/rate", True),
    ],
)
def test_compile_arn_regex(expression, resource, match):
    regex = RBACAuthorizationBackend.compile_arn_regex(expression)
    assert (regex.match(resource) is not None) == match


@pytest.mark.parametrize(
    "expression, namespace, name",
    [
        ("e*/d*", "e%", "d%"),
        ("*e*d*/*e", "%e%d%", "%e"),
    ],
)
def test_compile_arn_sql_like(expression, namespace, name):
    result = RBACAuthorizationBackend.compile_arn_sql_like(expression)
    assert (namespace, name) == result


@pytest.mark.parametrize(
    "token_string,authenticated",
    [
        ("", False),
        ("asdf", False),
    ],
)
def test_authenticated(token_string, authenticated):
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"
    assert (authentication.authenticate(token_string) is not None) == authenticated


def test_valid_token():
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace="default",
            role_bindings={
                "default/*": ["viewer"],
                "e*/e*": ["admin"],
            },
        )
    )

    token_model = authentication.authenticate(token)
    assert isinstance(token_model, AuthenticationToken)


def test_expired_token():
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace="default",
            exp=datetime.datetime.utcnow() - datetime.timedelta(hours=1),
            role_bindings={
                "default/*": ["viewer"],
                "e*/e*": ["admin"],
            },
        )
    )

    assert authentication.authenticate(token) is None


@pytest.mark.parametrize(
    "entity_bindings,arn,permissions,authorized",
    [
        (
            {
                "example-namespace/example-name": {"developer"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_CREATE},
            True,
        ),
        (
            {
                "example-namespace/example-name": {"developer", "viewer"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_DELETE},
            False,
        ),
        (
            {
                "example-namespace/example-name": {"developer", "admin"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_DELETE},
            True,
        ),
        (
            {
                "e*/e*am*": {"admin"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_CREATE},
            True,
        ),
        (
            {
                "e*/e*am*": {"viewer"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_CREATE},
            False,
        ),
    ],
)
def test_authorization(entity_bindings, arn, permissions, authorized):
    authorization = RBACAuthorizationBackend()
    assert authorized == authorization.authorize(entity_bindings, arn, permissions)


def test_end_to_end_auth_flow():
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace="default",
            role_bindings={
                "default/*": ["viewer"],
                "e*/e*": ["admin"],
            },
        )
    )

    token_model = authentication.authenticate(token)

    authorization = RBACAuthorizationBackend()
    assert authorization.authorize(
        token_model.role_bindings,
        "example-namespace/example-name",
        {
            Permissions.ENVIRONMENT_DELETE,
            Permissions.ENVIRONMENT_READ,
        },
        authenticated=True,
    )


@pytest.mark.parametrize(
    "arn_1,arn_2,value",
    [
        ("ab/cd", "a*b/c*d", True),
        ("a1111b/c22222d", "a*b/c*d", True),
        ("a1/cd", "a*b/c*d", False),
        ("abc/ed", "a*b*c/e*d", True),
        ("a111bc/ed", "a*b*c/e*d", True),
        ("a111b2222c/e3333d", "a*b*c/e*d", True),
        ("aaabbbcccc/eeddd", "a*b*c/e*d", True),
        ("aaabbbcccc1/eeddd", "a*b*c/e*d", False),
        ("aaabbbcccc1c/eeddd", "a*b*c/e*d", True),
    ],
)
def test_is_arn_subset(arn_1, arn_2, value):
    assert RBACAuthorizationBackend.is_arn_subset(arn_1, arn_2) == value


@pytest.mark.parametrize(
    "entity_bindings, new_entity_bindings, authenticated, value",
    [
        # */* viewer is a subset of admin
        ({"*/*": ["admin"]}, {"*/*": ["viewer"]}, False, True),
        ({"*/*": ["admin"]}, {"*/*": ["viewer"]}, True, True),
        # */* admin is not a subset of viewer
        ({"*/*": ["viewer"]}, {"*/*": ["admin"]}, False, False),
        ({"*/*": ["viewer"]}, {"*/*": ["admin"]}, True, False),
        # a/b viewer is a subset of admin
        ({"a/b": ["admin"]}, {"a/b": ["viewer"]}, False, True),
        ({"a/b": ["admin"]}, {"a/b": ["viewer"]}, True, True),
        # a/b admin is not a subset of viewer
        ({"a/b": ["viewer"]}, {"a/b": ["admin"]}, False, False),
        ({"a/b": ["viewer"]}, {"a/b": ["admin"]}, True, False),
        # default/* vs. */*
        ({"*/*": ["viewer"]}, {"default/*": ["viewer"]}, False, True),
        ({"*/*": ["viewer"]}, {"default/*": ["viewer"]}, True, True),
        # efault/* vs. d*/*
        ({"d*/*": ["viewer"]}, {"efault/*": ["viewer"]}, False, False),
        ({"d*/*": ["viewer"]}, {"efault/*": ["viewer"]}, True, False),
        # multiple entities keys
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"default/*": ["developer"]},
            False,
            True,
        ),
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"default/*": ["developer"]},
            True,
            True,
        ),
        # multiple entities keys
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"dcefault/*": ["developer"]},
            False,
            False,
        ),
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"dcefault/*": ["developer"]},
            True,
            False,
        ),
        # multiple entities keys
        ({"d*/*": ["viewer"]}, {"d*/*": ["viewer"], "dc*/*": ["viewer"]}, False, True),
        ({"d*/*": ["viewer"]}, {"d*/*": ["viewer"], "dc*/*": ["viewer"]}, True, True),
        # multiple entities keys
        (
            {"d*/*": ["viewer"]},
            {"d*/*": ["viewer"], "dc*/*": ["developer"]},
            False,
            False,
        ),
        (
            {"d*/*": ["viewer"]},
            {"d*/*": ["viewer"], "dc*/*": ["developer"]},
            True,
            False,
        ),
    ],
)
def test_is_subset_entity_permissions(
    entity_bindings, new_entity_bindings, authenticated, value
):
    authorization = RBACAuthorizationBackend()
    assert (
        authorization.is_subset_entity_permissions(
            entity_bindings, new_entity_bindings, authenticated
        )
        == value
    )



import os
import datetime
import shutil

from sqlalchemy import (
    Table,
    Column,
    BigInteger,
    Integer,
    Unicode,
    Text,
    JSON,
    Enum,
    DateTime,
    UniqueConstraint,
    ForeignKey,
    UnicodeText,
    or_,
    and_,
)
from sqlalchemy.orm import sessionmaker, relationship, scoped_session, backref
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import create_engine

from conda_store_server import utils, schema
from conda_store_server.environment import validate_environment
from conda_store_server.conda import download_repodata
import logging

logger = logging.getLogger("orm")

Base = declarative_base()


class Namespace(Base):
    """Namespace for resources"""

    __tablename__ = "namespace"

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(255), unique=True)

    environments = relationship("Environment", back_populates="namespace")

    deleted_on = Column(DateTime, default=None)


class Specification(Base):
    """The specifiction for a given conda environment"""

    __tablename__ = "specification"

    def __init__(self, specification):
        if not validate_environment(specification):
            raise ValueError(
                "specification={specification} is not valid conda environment.yaml"
            )
        self.name = specification["name"]
        self.spec = specification
        self.sha256 = utils.datastructure_hash(self.spec)

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(255), nullable=False)
    spec = Column(JSON, nullable=False)
    sha256 = Column(Unicode(255), unique=True, nullable=False)
    created_on = Column(DateTime, default=datetime.datetime.utcnow)

    builds = relationship("Build", back_populates="specification")
    solves = relationship("Solve", back_populates="specification")


solve_conda_package_build = Table(
    "solve_conda_package_build",
    Base.metadata,
    Column("solve_id", ForeignKey("solve.id", ondelete="CASCADE"), primary_key=True),
    Column(
        "conda_package_build_id",
        ForeignKey("conda_package_build.id", ondelete="CASCADE"),
        primary_key=True,
    ),
)


class Solve(Base):
    """A solve for a particular specification"""

    __tablename__ = "solve"

    id = Column(Integer, primary_key=True)
    specification_id = Column(Integer, ForeignKey("specification.id"), nullable=False)
    specification = relationship(Specification, back_populates="solves")

    scheduled_on = Column(DateTime, default=datetime.datetime.utcnow)
    started_on = Column(DateTime, default=None)
    ended_on = Column(DateTime, default=None)

    package_builds = relationship(
        "CondaPackageBuild", secondary=solve_conda_package_build
    )


build_conda_package = Table(
    "build_conda_package_build",
    Base.metadata,
    Column("build_id", ForeignKey("build.id", ondelete="CASCADE"), primary_key=True),
    Column(
        "conda_package_build_id",
        ForeignKey("conda_package_build.id", ondelete="CASCADE"),
        primary_key=True,
    ),
)


class Build(Base):
    """The state of a build of a given specification"""

    __tablename__ = "build"

    id = Column(Integer, primary_key=True)
    specification_id = Column(Integer, ForeignKey("specification.id"), nullable=False)
    specification = relationship(Specification, back_populates="builds")

    environment_id = Column(Integer, ForeignKey("environment.id"), nullable=False)
    environment = relationship(
        "Environment",
        backref=backref("builds", cascade="all, delete-orphan"),
        foreign_keys=[environment_id],
    )

    package_builds = relationship("CondaPackageBuild", secondary=build_conda_package)

    status = Column(Enum(schema.BuildStatus), default=schema.BuildStatus.QUEUED)
    size = Column(BigInteger, default=0)
    scheduled_on = Column(DateTime, default=datetime.datetime.utcnow)
    started_on = Column(DateTime, default=None)
    ended_on = Column(DateTime, default=None)
    deleted_on = Column(DateTime, default=None)

    build_artifacts = relationship(
        "BuildArtifact", back_populates="build", cascade="all, delete-orphan"
    )

    def build_path(self, conda_store):
        """Build path is the directory for the conda prefix used to
        build the environment

        """
        store_directory = os.path.abspath(conda_store.store_directory)
        namespace = self.environment.namespace.name
        name = self.specification.name
        return os.path.join(
            conda_store.build_directory.format(
                store_directory=store_directory, namespace=namespace, name=name
            ),
            self.build_key,
        )

    def environment_path(self, conda_store):
        """Environment path is the path for the symlink to the build
        path

        """
        store_directory = os.path.abspath(conda_store.store_directory)
        namespace = self.environment.namespace.name
        name = self.specification.name
        return conda_store.environment_directory.format(
            store_directory=store_directory, namespace=namespace, name=name
        )

    @property
    def build_key(self):
        """A conda environment build is a function of the sha256 of the
        environment.yaml along with the time that the package was built.

        The last two parts of the build key are to assist finding the
        record in the database.

        The build key should be a key that allows for the environment
        build to be easily identified and found in the database.
        """
        datetime_format = "%Y%m%d-%H%M%S-%f"
        return f"{self.specification.sha256}-{self.scheduled_on.strftime(datetime_format)}-{self.id}-{self.specification.name}"

    @staticmethod
    def parse_build_key(key):
        parts = key.split("-")
        if len(parts) < 5:
            return None
        return int(parts[4])  # build_id

    @property
    def log_key(self):
        return f"logs/{self.build_key}.log"

    @property
    def conda_env_export_key(self):
        return f"yaml/{self.build_key}.yml"

    @property
    def conda_pack_key(self):
        return f"archive/{self.build_key}.tar.gz"

    @property
    def docker_manifest_key(self):
        return f"docker/manifest/{self.build_key}"

    def docker_blob_key(self, blob_hash):
        return f"docker/blobs/{blob_hash}"

    @hybrid_property
    def has_lockfile(self):
        return any(
            artifact.artifact_type == schema.BuildArtifactType.LOCKFILE
            for artifact in self.build_artifacts
        )

    @hybrid_property
    def has_yaml(self):
        return any(
            artifact.artifact_type == schema.BuildArtifactType.YAML
            for artifact in self.build_artifacts
        )

    @hybrid_property
    def has_conda_pack(self):
        return any(
            artifact.artifact_type == schema.BuildArtifactType.CONDA_PACK
            for artifact in self.build_artifacts
        )

    @hybrid_property
    def has_docker_manifest(self):
        return any(
            artifact.artifact_type == schema.BuildArtifactType.DOCKER_MANIFEST
            for artifact in self.build_artifacts
        )

    def __repr__(self):
        return f"<Build (id={self.id} status={self.status} nbr package_builds={len(self.package_builds)})>"


class BuildArtifact(Base):
    """Artifacts of a given build"""

    __tablename__ = "build_artifact"

    id = Column(Integer, primary_key=True)

    build_id = Column(Integer, ForeignKey("build.id"))
    build = relationship(Build, back_populates="build_artifacts")

    artifact_type = Column(Enum(schema.BuildArtifactType), nullable=False)

    key = Column(Unicode(255))


class Environment(Base):
    """Pointer to the current build and specification for a given
    environment name

    """

    __tablename__ = "environment"

    __table_args__ = (
        UniqueConstraint("namespace_id", "name", name="_namespace_name_uc"),
    )

    id = Column(Integer, primary_key=True)

    namespace_id = Column(Integer, ForeignKey("namespace.id"), nullable=False)
    namespace = relationship(Namespace)

    name = Column(Unicode(255), nullable=False)

    current_build_id = Column(Integer, ForeignKey("build.id"))
    current_build = relationship(
        Build, foreign_keys=[current_build_id], post_update=True
    )

    deleted_on = Column(DateTime, default=None)

    description = Column(UnicodeText, default=None)


class CondaChannel(Base):
    __tablename__ = "conda_channel"

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(255), unique=True, nullable=False)
    last_update = Column(DateTime)

    def update_packages(self, db, subdirs=None):
        logger.info(f"update packages {self.name} ")

        logger.info("Downloading repodata ...  ")
        repodata = download_repodata(self.name, self.last_update, subdirs=subdirs)
        logger.info("repodata downloaded ")

        # Hint : you can store the file locally for later debug
        # with open(f"~/{self.name.split('/')[-1]}_{datetime.datetime.now()}.json", "w") as f:
        #    f.write(json.dumps(repodata, indent=4, sort_keys=True))

        if not repodata:
            # nothing to update
            return

        for architecture in repodata["architectures"]:
            logger.info(f"architecture  : {architecture} ")

            """
            Context :
               For each architecture, we need to add all the packages (`conda_package`),
               and all their builds (`conda_package_build`), with their relationship.
               As of May 2022, based on the default channels `main` and `conda-forge`,
               there are 136K packages and 367K conda_package_build

            Trick :
               To insert these data fast, we need to get rid of the overhead induced
               by the ORM layer, and use session.bulk_insert_mappings.

            Caveat :
               bulk insertion is handy, but we need to avoid breaking integrity constraint,
               This implies that we need to bulk_insert only new data, filtering out
               the data already in the DB

            Algorithm :
               First step :  we insert all the new conda_package rows
               Second step : we insert all the new conda_package_builds rows

               These steps are detailled below.
            """

            # package_data contains all the data for the iterated architecture.
            # Each dict represents a conda_package_build and also contains
            # thge data of the conda_package
            packages_data = list(
                repodata["architectures"][architecture]["packages"].values()
            )

            # First, we retrieve all the pairs "package name - pacakge version"
            # in the DB. This represents all the existing packages.
            existing_packages_keys = [
                f"{_[0]}-{_[1]}-{self.id}"
                for _ in db.query(CondaPackage.name, CondaPackage.version)
                .filter(CondaPackage.channel_id == self.id)
                .all()
            ]

            # Then, we filter packages_data to keep only the new packages.

            # `packages` associates a key representing the package like "{name}-{version}-{channel_id}"
            # to a dict representing a new package to insert.
            # By using a dict with such key, we avoid potential duplicates from within the repodata.
            packages = {}

            for p_build in packages_data:

                package_key = f'{p_build["name"]}-{p_build["version"]}-{self.id}'

                # Filtering out : if the key already exists in existing_packages_keys,
                # then the package is already if DB, we don't add it.
                if (
                    package_key not in packages
                    and package_key not in existing_packages_keys
                ):

                    new_package_dict = {
                        "channel_id": self.id,
                        "license": p_build.get("license"),
                        "license_family": p_build.get("license_family"),
                        "name": p_build["name"],
                        "version": p_build["version"],
                        "summary": repodata.get("packages", {})
                        .get(p_build["name"], {})
                        .get("summary"),
                        "description": repodata.get("packages", {})
                        .get(p_build["name"], {})
                        .get("description"),
                    }

                    packages[package_key] = new_package_dict

            logger.info(f"packages to insert : {len(packages)} ")

            try:
                db.bulk_insert_mappings(CondaPackage, packages.values())
                db.commit()
            except Exception as e:
                print(f"{e}")
                db.rollback()
                raise e

            logger.info("insert packages done")

            logger.info("retrieving existing sha256  : ")
            existing_sha256 = {
                _[0]
                for _ in db.query(CondaPackageBuild.sha256)
                .join(CondaPackageBuild.package)
                .filter(CondaPackage.channel_id == self.id)
                .filter(CondaPackageBuild.subdir == architecture)
                .all()
            }
            logger.info("retrieved existing sha256  : ")
            logger.info(f"package data before filtering  : {len(packages_data)} ")

            # We store the package builds in a dict indexed by their sha256
            # Later, we keep only the values.
            # That way, any duplicated sha256 from the repodata is erased
            # Also, we exclude pacakge_builds for which the sha256 is already in the DB
            packages_builds = {}
            for pb in packages_data:
                if pb["sha256"] not in existing_sha256:
                    packages_builds[pb["sha256"]] = pb

            packages_builds = packages_builds.values()
            logger.info(f"package builds after filtering : {len(packages_builds)} ")

            # This associates a tuple like "(name,version)" representing a package
            # to all its builds
            package_builds = {}
            logger.info("Creating CondaPackageBuild objects")
            for p_build in packages_builds:

                has_null = False
                non_null_keys = [
                    "build",
                    "build_number",
                    "depends",
                    "md5",
                    "sha256",
                    "size",
                ]
                for k in non_null_keys:
                    if p_build[k] is None:
                        has_null = True
                        break
                if has_null:
                    continue

                if p_build["depends"] == []:
                    p_build["depends"] = ""

                package_key = (p_build["name"], p_build["version"])

                new_package_build_dict = {
                    "build": p_build["build"],
                    "build_number": p_build["build_number"],
                    "channel_id": self.id,
                    "constrains": p_build.get("constrains"),
                    "depends": p_build["depends"],
                    "md5": p_build["md5"],
                    "sha256": p_build["sha256"],
                    "size": p_build["size"],
                    "subdir": p_build.get("subdir"),
                    "timestamp": p_build.get("timestamp"),
                }

                if package_key not in package_builds:
                    package_builds[package_key] = []

                package_builds[package_key].append(new_package_build_dict)
            logger.info("CondaPackageBuild objects created")

            batch_size = 1000
            all_package_keys = list(package_builds.keys())
            for i in range(0, len(all_package_keys), batch_size):

                logger.info(f"handling subset at index {i} (batch size {batch_size}")
                subset_keys = all_package_keys[i : i + batch_size]

                # retrieve the parent packages for the subset
                logger.info("retrieve the parent packages for the subset ")
                statements = []
                for p_name, p_version in subset_keys:
                    statements.append(
                        and_(
                            CondaPackage.name == p_name,
                            CondaPackage.version == p_version,
                        )
                    )
                all_parent_packages = (
                    db.query(CondaPackage).filter(or_(*statements)).all()
                )
                all_parent_packages = {
                    (_.name, _.version): _ for _ in all_parent_packages
                }
                logger.info(f"parent packages retrieved : {len(all_parent_packages)} ")

                for p_name, p_version in subset_keys:
                    for p_build_dict in package_builds[(p_name, p_version)]:
                        p_build_dict["package_id"] = all_parent_packages[
                            (p_name, p_version)
                        ].id

                logger.info("ready to bulk save")

            try:

                flatten = []
                for (p_name, p_version) in package_builds:
                    flatten += package_builds[(p_name, p_version)]

                db.bulk_insert_mappings(CondaPackageBuild, flatten)
                db.commit()
                logger.info("bulk saved")
            except Exception as e:
                logger.error(f"{e}")

                raise e

            logger.info(f"DONE for architecture  : {architecture}")

        self.last_update = datetime.datetime.utcnow()
        db.commit()
        logger.info("update packages DONE ")


class CondaPackage(Base):
    __tablename__ = "conda_package"

    __table_args__ = (
        UniqueConstraint(
            "channel_id",
            "name",
            "version",
            name="_conda_package_uc",
        ),
    )

    id = Column(Integer, primary_key=True)

    channel_id = Column(Integer, ForeignKey("conda_channel.id"), index=True)
    channel = relationship(CondaChannel)

    builds = relationship(
        "CondaPackageBuild", back_populates="package", cascade="all, delete-orphan"
    )

    license = Column(Text, nullable=True)
    license_family = Column(Unicode(64), nullable=True)
    name = Column(Unicode(255), nullable=False, index=True)
    version = Column(Unicode(64), nullable=False, index=True)
    summary = Column(Text, nullable=True)
    description = Column(Text, nullable=True)

    def __repr__(self):
        return f"<CondaPackage (channel={self.channel} name={self.name} version={self.version})>"


class CondaPackageBuild(Base):
    __tablename__ = "conda_package_build"

    __table_args__ = (
        UniqueConstraint(
            "channel_id",
            "package_id",
            "subdir",
            "build",
            "build_number",
            "sha256",
            name="_conda_package_build_uc",
        ),
    )

    id = Column(Integer, primary_key=True)

    package_id = Column(Integer, ForeignKey("conda_package.id"))
    package = relationship(CondaPackage, back_populates="builds")

    """
    Some package builds have the exact same data from different channels.
    Thus, when adding a channel, populating CondaPackageBuild can encounter
    duplicate keys errors. That's why we need to distinguish them by channel_id.
    """
    channel_id = Column(Integer, ForeignKey("conda_channel.id"))
    channel = relationship(CondaChannel)

    build = Column(Unicode(64), nullable=False, index=True)
    build_number = Column(Integer, nullable=False)
    constrains = Column(JSON, nullable=True)
    depends = Column(JSON, nullable=False)
    md5 = Column(Unicode(255), nullable=False)
    sha256 = Column(Unicode(64), nullable=False)
    size = Column(BigInteger, nullable=False)
    subdir = Column(Unicode(64), nullable=True)
    timestamp = Column(BigInteger, nullable=True)

    def __repr__(self):
        return f"<CondaPackageBuild (id={self.id} build={self.build} size={self.size} sha256={self.sha256})>"


class CondaStoreConfiguration(Base):
    __tablename__ = "conda_store_configuration"

    id = Column(Integer, primary_key=True)

    disk_usage = Column(BigInteger, default=0)
    free_storage = Column(BigInteger, default=0)
    total_storage = Column(BigInteger, default=0)

    @classmethod
    def configuration(cls, db):
        query = db.query(cls).filter(cls.id == 1)
        if query.count() == 0:
            db.add(cls(id=1))
            db.commit()
        return query.first()

    @classmethod
    def update_storage_metrics(cls, db, store_directory):
        configuration = cls.configuration(db)

        disk_usage = shutil.disk_usage(store_directory)
        configuration.disk_usage = disk_usage.used
        configuration.free_storage = disk_usage.free
        configuration.total_storage = disk_usage.total
        db.commit()


def new_session_factory(url="sqlite:///:memory:", reset=False, **kwargs):
    engine = create_engine(url, **kwargs)

    session_factory = scoped_session(sessionmaker(bind=engine))
    return session_factory



"""Add Environment.description

Revision ID: 8d63a091aff8
Revises: 48be4072fe58
Create Date: 2022-07-15 14:22:00.351131

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "8d63a091aff8"
down_revision = "48be4072fe58"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("environment") as batch_op:
        batch_op.add_column(sa.Column("description", sa.UnicodeText(), nullable=True))


def downgrade():
    with op.batch_alter_table("environment") as batch_op:
        batch_op.drop_column("description")



from conda_store_server.server.views.api import router_api  # noqa
from conda_store_server.server.views.registry import router_registry  # noqa
from conda_store_server.server.views.ui import router_ui  # noqa
from conda_store_server.server.views.metrics import router_metrics  # noqa
from conda_store_server.server.views.conda_store_ui import router_conda_store_ui  # noqa



from fastapi import APIRouter, Depends
from fastapi.responses import PlainTextResponse

from conda_store_server import api
from conda_store_server.server import dependencies


router_metrics = APIRouter(tags=["metrics"])


@router_metrics.get("/metrics", response_class=PlainTextResponse)
def prometheus_metrics(
    conda_store=Depends(dependencies.get_conda_store),
):
    metrics = api.get_metrics(conda_store.db)
    return "\n".join(f"conda_store_{key} {value}" for key, value in metrics.items())


@router_metrics.get("/celery")
def trigger_task(conda_store=Depends(dependencies.get_conda_store)):
    conda_store.celery_app

    def get_celery_worker_status(app):
        i = app.control.inspect()
        availability = i.ping()
        stats = i.stats()
        registered_tasks = i.registered()
        active_tasks = i.active()
        scheduled_tasks = i.scheduled()
        result = {
            "availability": availability,
            "stats": stats,
            "registered_tasks": registered_tasks,
            "active_tasks": active_tasks,
            "scheduled_tasks": scheduled_tasks,
        }
        return result

    return get_celery_worker_status(conda_store.celery_app)



"""Interface between conda-store and conda

This module provides all the functionality that is required for
executing conda commands
"""

import os
import json
import subprocess
import bz2
import datetime
import hashlib
import tempfile
import pathlib

import yaml
import yarl
import requests

from conda_store_server import schema


def normalize_channel_name(channel_alias, channel):
    if channel.startswith("http"):
        return channel

    channel_alias = yarl.URL(channel_alias)
    return str(channel_alias / channel)


def conda_list(prefix, executable: str = "conda"):
    args = [executable, "list", "-p", prefix, "--json"]
    return json.loads(subprocess.check_output(args))


def conda_pack(prefix, output, ignore_missing_files=True):
    from conda_pack import pack

    pack(
        prefix=str(prefix),
        output=str(output),
        ignore_missing_files=ignore_missing_files,
    )


def conda_lock(specification: schema.CondaSpecification, conda_exe: str = "mamba"):
    from conda_lock.conda_lock import run_lock
    from conda.models.dist import Dist

    with tempfile.TemporaryDirectory() as tmpdir:
        environment_path = pathlib.Path(tmpdir) / "environment.yaml"
        lockfile_path = pathlib.Path(tmpdir) / "environment-lock.yaml"

        with environment_path.open("w") as f:
            f.write(specification.json())

        try:
            run_lock(
                environment_files=[environment_path],
                platforms=[conda_platform()],
                lockfile_path=lockfile_path,
                conda_exe=conda_exe,
            )
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output)

        with lockfile_path.open() as f:
            lockfile = yaml.safe_load(f)

    conda_packages = []
    pip_packages = []

    for package in lockfile["package"]:
        if package["manager"] == "conda":
            dist = Dist.from_string(package["url"])
            conda_packages.append(
                {
                    "name": dist.name,
                    "build": dist.build,
                    "build_number": dist.build_number,
                    "constrains": None,
                    "depends": [],
                    "license": None,
                    "license_family": None,
                    "size": -1,
                    "subdir": dist.subdir,
                    "timestamp": None,
                    "version": dist.version,
                    "channel_id": dist.base_url,
                    "md5": package["hash"].get("md5"),
                    "sha256": package["hash"].get("sha256", ""),
                    "summary": None,
                    "description": None,
                }
            )
        elif package["manager"] == "pip":
            pip_packages.append(
                {
                    "name": package["name"],
                    "url": package["url"],
                    "version": package["version"],
                    "sha256": package["hash"]["sha256"],
                }
            )

    return {"conda": conda_packages, "pip": pip_packages}


def download_repodata(
    channel: str,
    last_update: datetime.datetime = None,
    subdirs=None,
):
    """Download repodata for channel only if changed since last update

    A channel consists of several architectures: linux-32, linux-64,
    linux-aarch64, linux-armv6l, linux-armv7l, linux-ppc64le, noarch,
    osx-64, win-32, win-64, zos-z (possibly others).

    Check ``conda.base.constants.KNOWN_SUBDIRS`` for the full list.
    """
    subdirs = set(subdirs or [conda_platform(), "noarch"])

    # the conda main channel does not have a channeldata.json within
    # the https://conda.anaconda.org/<channel>/channeldata.json
    # so we replace the given url with it's equivalent alias
    channel_replacements = {
        "https://conda.anaconda.org/main/": yarl.URL(
            "https://repo.anaconda.com/pkgs/main"
        )
    }
    normalized_channel_url = yarl.URL(channel) / "./"
    channel_url = channel_replacements.get(
        str(normalized_channel_url), yarl.URL(channel)
    )

    headers = {}
    if last_update:
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Modified-Since
        # only download channeldata.json if it has been modified since
        # last update to minimize bandwidth usage of channel updates
        # most static web servers obey this header otherwise it will
        # be ignored
        headers["If-Modified-Since"] = last_update.strftime("%a, %d %b %Y %H:%M:%S GMT")

    response = requests.get(channel_url / "channeldata.json", headers=headers)
    if response.status_code == 304:  # 304 Not Modified since last_update
        return {"architectures": {}}
    response.raise_for_status()

    repodata = response.json()
    repodata["architectures"] = {}
    for subdir in subdirs:
        response = requests.get(
            channel_url / subdir / "repodata.json.bz2", headers=headers
        )
        if response.status_code == 304:  # 304 Not Modified since last_update
            continue
        response.raise_for_status()
        repodata["architectures"][subdir] = json.loads(bz2.decompress(response.content))
    return repodata


def conda_platform():
    """
    Return the platform (in ``conda`` style) the server is running on.

    It will be one of the values in ``conda.base.constants.KNOWN_SUBDIRS``.
    """
    from conda.base.context import context

    return context.subdir


def conda_prefix_packages(prefix):
    """
    Returns a list of the packages that exist for a given prefix

    """
    from conda.core.prefix_data import PrefixData

    packages = []

    prefix_data = PrefixData(prefix)
    prefix_data.load()

    for record in prefix_data.iter_records():
        package = {
            "build": record.build,
            "build_number": record.build_number,
            "constrains": list(record.constrains),
            "depends": list(record.depends),
            "license": record.license,
            "license_family": record.license_family,
            "md5": hashlib.md5(
                open(record.package_tarball_full_path, "rb").read()
            ).hexdigest(),
            "sha256": hashlib.sha256(
                open(record.package_tarball_full_path, "rb").read()
            ).hexdigest(),
            "name": record.name,
            "size": record.size,
            "subdir": record.subdir,
            "timestamp": record.timestamp,
            "version": record.version,
            "channel_id": record.channel.base_url,
            "summary": None,
            "description": None,
        }

        info_json = os.path.join(record.extracted_package_dir, "info/about.json")
        if os.path.exists(info_json):
            info = json.load(open(info_json))
            package["summary"] = info.get("summary")
            package["description"] = info.get("description")

        packages.append(package)
    return packages



from fastapi import APIRouter, Request, Depends

from conda_store_server.server import dependencies

router_conda_store_ui = APIRouter(tags=["conda-store-ui"])


@router_conda_store_ui.get("/")
def get_conda_store_ui(
    request: Request,
    templates=Depends(dependencies.get_templates),
):
    context = {
        "request": request,
    }
    return templates.TemplateResponse("conda-store-ui.html", context)
